# SPDX-License-Identifier: GPL-3.0-or-later

# https://extensions.blender.org/add-ons/c3db/

import bpy
from mathutils import Quaternion
from mathutils import Vector


class C3DB_PG_properties(bpy.types.PropertyGroup):
    location: bpy.props.FloatVectorProperty(name="Location")
    rotation_quaternion: bpy.props.FloatVectorProperty(
        name="Rotation Quaternion", size=4
    )
    rotation_axis_angle: bpy.props.FloatVectorProperty(
        name="Rotation Axis Angle", size=4
    )
    rotation_euler: bpy.props.FloatVectorProperty(name="Rotation euler", size=3)
    rotation_mode: bpy.props.StringProperty(name="Rotation Mode")


class C3DB_MT_menu_specials(bpy.types.Menu):
    bl_label = "3D Cursor Specials"
    bl_idname = "C3DB_MT_menu_specials"

    def draw(self, context):
        layout = self.layout
        layout.operator("c3db.restore_previous", icon="TRIA_UP_BAR")
        layout.operator("c3db.restore_next", icon="TRIA_DOWN_BAR")
        layout.separator()
        layout.operator("c3db.update", icon="FILE_REFRESH")
        layout.separator()
        layout.operator("c3db.delete_all", icon="TRASH")
        layout.separator()
        layout.operator("c3db.convert_selected_to_empty", icon="OUTLINER")
        layout.operator("c3db.convert_all_to_empties")


class C3DB_UL_list(bpy.types.UIList):
    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        layout.prop(item, "name", text="", emboss=False)


class C3DB_OT_save(bpy.types.Operator):
    """Save 3D Cursor location, rotation and rotation mode"""

    bl_idname = "c3db.save"
    bl_label = "Save 3D Cursor"

    def execute(self, context):
        item = context.scene.C3DB_3Dcursors_collection.add()
        item.name = "3D Cursor"
        item.location = context.scene.cursor.location
        item.rotation_mode = context.scene.cursor.rotation_mode
        if item.rotation_mode == "QUATERNION":
            item.rotation_quaternion = context.scene.cursor.rotation_quaternion
        elif item.rotation_mode == "AXIS_ANGLE":
            item.rotation_axis_angle = context.scene.cursor.rotation_axis_angle
        else:
            item.rotation_euler = context.scene.cursor.rotation_euler
        # Select the 3D cursor currently saving from the list
        context.scene.C3DB_3Dcursors_index = (
            len(context.scene.C3DB_3Dcursors_collection) - 1
        )
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_restore(bpy.types.Operator):
    """Restore selected 3D Cursor"""

    bl_idname = "c3db.restore"
    bl_label = "Load 3D Cursor"

    @classmethod
    def poll(cls, context):
        return bool(
            context.scene.C3DB_3Dcursors_collection
        ) and context.scene.C3DB_3Dcursors_index < len(
            context.scene.C3DB_3Dcursors_collection
        )

    def execute(self, context):
        if context.scene.C3DB_3Dcursors_collection:
            item = context.scene.C3DB_3Dcursors_collection[
                context.scene.C3DB_3Dcursors_index
            ]
            context.scene.cursor.location = item.location
            context.scene.cursor.rotation_mode = item.rotation_mode
            if item.rotation_mode == "QUATERNION":
                context.scene.cursor.rotation_quaternion = item.rotation_quaternion
            elif item.rotation_mode == "AXIS_ANGLE":
                context.scene.cursor.rotation_axis_angle = item.rotation_axis_angle
            else:
                context.scene.cursor.rotation_euler = item.rotation_euler
            if context.scene.c3db_auto_center_view_checkbox:
                bpy.ops.view3d.view_center_cursor()
            context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_update(bpy.types.Operator):
    """Update selected 3D Cursor"""

    bl_idname = "c3db.update"
    bl_label = "Update 3D Cursor"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return bool(
            context.scene.C3DB_3Dcursors_collection
        ) and context.scene.C3DB_3Dcursors_index < len(
            context.scene.C3DB_3Dcursors_collection
        )

    def execute(self, context):
        item = context.scene.C3DB_3Dcursors_collection[
            context.scene.C3DB_3Dcursors_index
        ]
        item.location = context.scene.cursor.location
        item.rotation_mode = context.scene.cursor.rotation_mode
        if item.rotation_mode == "QUATERNION":
            item.rotation_quaternion = context.scene.cursor.rotation_quaternion
        elif item.rotation_mode == "AXIS_ANGLE":
            item.rotation_axis_angle = context.scene.cursor.rotation_axis_angle
        else:
            item.rotation_euler = context.scene.cursor.rotation_euler
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_restore_next(bpy.types.Operator):
    """Go to next 3D Cursor and Restore it"""

    bl_idname = "c3db.restore_next"
    bl_label = "Load next 3D Cursor"

    @classmethod
    def poll(cls, context):
        return (
            bool(context.scene.C3DB_3Dcursors_collection)
            and len(context.scene.C3DB_3Dcursors_collection) > 1
        )

    def execute(self, context):
        # Select the next valid 3D cursor in the list
        C3DB_3Dcursors_collection = context.scene.C3DB_3Dcursors_collection
        C3DB_3Dcursors_index = context.scene.C3DB_3Dcursors_index
        if C3DB_3Dcursors_index < len(C3DB_3Dcursors_collection) - 1:
            context.scene.C3DB_3Dcursors_index = C3DB_3Dcursors_index + 1
        elif C3DB_3Dcursors_index == len(C3DB_3Dcursors_collection) - 1:
            context.scene.C3DB_3Dcursors_index = 0
        # Restore selected 3D cursor in the list
        if context.scene.C3DB_3Dcursors_collection:
            item = context.scene.C3DB_3Dcursors_collection[
                context.scene.C3DB_3Dcursors_index
            ]
            context.scene.cursor.location = item.location
            context.scene.cursor.rotation_mode = item.rotation_mode
            if item.rotation_mode == "QUATERNION":
                context.scene.cursor.rotation_quaternion = item.rotation_quaternion
            elif item.rotation_mode == "AXIS_ANGLE":
                context.scene.cursor.rotation_axis_angle = item.rotation_axis_angle
            else:
                context.scene.cursor.rotation_euler = item.rotation_euler
        if context.scene.c3db_auto_center_view_checkbox:
            bpy.ops.view3d.view_center_cursor()
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_restore_previous(bpy.types.Operator):
    """Go to previous 3D Cursor and Restore it"""

    bl_idname = "c3db.restore_previous"
    bl_label = "Load previous 3D Cursor"

    @classmethod
    def poll(cls, context):
        return (
            bool(context.scene.C3DB_3Dcursors_collection)
            and len(context.scene.C3DB_3Dcursors_collection) > 1
        )

    def execute(self, context):
        # Select the previous valid 3D cursor in the list
        C3DB_3Dcursors_collection = context.scene.C3DB_3Dcursors_collection
        C3DB_3Dcursors_index = context.scene.C3DB_3Dcursors_index
        if C3DB_3Dcursors_index > 0:
            context.scene.C3DB_3Dcursors_index = C3DB_3Dcursors_index - 1
        elif C3DB_3Dcursors_index == 0:
            context.scene.C3DB_3Dcursors_index = len(C3DB_3Dcursors_collection) - 1
        # Restore selected 3D cursor in the list
        if context.scene.C3DB_3Dcursors_collection:
            item = context.scene.C3DB_3Dcursors_collection[
                context.scene.C3DB_3Dcursors_index
            ]
            context.scene.cursor.location = item.location
            context.scene.cursor.rotation_mode = item.rotation_mode
            if item.rotation_mode == "QUATERNION":
                context.scene.cursor.rotation_quaternion = item.rotation_quaternion
            elif item.rotation_mode == "AXIS_ANGLE":
                context.scene.cursor.rotation_axis_angle = item.rotation_axis_angle
            else:
                context.scene.cursor.rotation_euler = item.rotation_euler
        if context.scene.c3db_auto_center_view_checkbox:
            bpy.ops.view3d.view_center_cursor()
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_delete_all(bpy.types.Operator):
    """Delete all 3D Cursors"""

    bl_idname = "c3db.delete_all"
    bl_label = "Delete All 3D Cursors"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return bool(
            context.scene.C3DB_3Dcursors_collection
        ) and context.scene.C3DB_3Dcursors_index < len(
            context.scene.C3DB_3Dcursors_collection
        )

    def execute(self, context):
        context.scene.C3DB_3Dcursors_collection.clear()
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_remove_from_list(bpy.types.Operator):
    """Delete selected 3D Cursor"""

    bl_idname = "c3db.remove_from_list"
    bl_label = "Delete 3D Cursor"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return bool(
            context.scene.C3DB_3Dcursors_collection
        ) and context.scene.C3DB_3Dcursors_index < len(
            context.scene.C3DB_3Dcursors_collection
        )

    def execute(self, context):
        C3DB_3Dcursors_collection = context.scene.C3DB_3Dcursors_collection
        C3DB_3Dcursors_index = context.scene.C3DB_3Dcursors_index

        C3DB_3Dcursors_collection.remove(C3DB_3Dcursors_index)

        # Select the next valid 3D cursor in the list
        if C3DB_3Dcursors_index < len(C3DB_3Dcursors_collection):
            context.scene.C3DB_3Dcursors_index = C3DB_3Dcursors_index
        elif C3DB_3Dcursors_collection:
            context.scene.C3DB_3Dcursors_index = len(C3DB_3Dcursors_collection) - 1
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_move_up_list(bpy.types.Operator):
    """Move the selected 3D Cursor up in the list"""

    bl_idname = "c3db.move_up_list"
    bl_label = "Move 3D Cursor Up the list"

    @classmethod
    def poll(cls, context):
        return context.scene.C3DB_3Dcursors_index > 0

    def execute(self, context):
        index = context.scene.C3DB_3Dcursors_index
        collection = context.scene.C3DB_3Dcursors_collection
        collection.move(index, index - 1)
        context.scene.C3DB_3Dcursors_index -= 1
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_move_down_list(bpy.types.Operator):
    """Move the selected 3D Cursor down in the list"""

    bl_idname = "c3db.move_down_list"
    bl_label = "Move 3D Cursor Down the list"

    @classmethod
    def poll(cls, context):
        return (
            context.scene.C3DB_3Dcursors_index
            < len(context.scene.C3DB_3Dcursors_collection) - 1
        )

    def execute(self, context):
        index = context.scene.C3DB_3Dcursors_index
        collection = context.scene.C3DB_3Dcursors_collection
        collection.move(index, index + 1)
        context.scene.C3DB_3Dcursors_index += 1
        context.area.tag_redraw()  # Refresh the UI
        return {"FINISHED"}


class C3DB_OT_convert_all_to_empties(bpy.types.Operator):
    """Convert all 3D Cursors to empty objects. This operator is inactive if: You are not in Object Mode ..."""

    bl_idname = "c3db.convert_all_to_empties"
    bl_label = "Convert All 3D Cursors to Empties"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return (
            bool(context.scene.C3DB_3Dcursors_collection)
            and context.scene.C3DB_3Dcursors_index
            < len(context.scene.C3DB_3Dcursors_collection)
            and context.mode == "OBJECT"
        )

    def execute(self, context):
        for c3db_3d_cursor in context.scene.C3DB_3Dcursors_collection:
            if c3db_3d_cursor.rotation_mode == "QUATERNION":
                rotation = Quaternion(c3db_3d_cursor.rotation_quaternion)
                rotation_euler = rotation.to_euler()
            elif c3db_3d_cursor.rotation_mode == "AXIS_ANGLE":
                rotation = Quaternion(
                    Vector(c3db_3d_cursor.rotation_axis_angle).yzw,
                    c3db_3d_cursor.rotation_axis_angle[0],
                )
                rotation_euler = rotation.to_euler()
            else:
                rotation_euler = c3db_3d_cursor.rotation_euler

            bpy.ops.object.empty_add(
                type="PLAIN_AXES",
                radius=1,
                location=c3db_3d_cursor.location,
                rotation=rotation_euler,
            )
            
            bpy.context.object.name = c3db_3d_cursor.name

        return {"FINISHED"}


class C3DB_OT_convert_selected_to_empty(bpy.types.Operator):
    """Convert selected 3D Cursor to an empty object. This operator is inactive if: You are not in Object Mode ..."""

    bl_idname = "c3db.convert_selected_to_empty"
    bl_label = "Convert Selected 3D Cursor to Empty"
    bl_options = {"REGISTER", "UNDO"}

    @classmethod
    def poll(cls, context):
        return (
            context.scene.C3DB_3Dcursors_collection
            and context.scene.C3DB_3Dcursors_index
            < len(context.scene.C3DB_3Dcursors_collection)
            and context.mode == "OBJECT"
        )

    def execute(self, context):
        item = context.scene.C3DB_3Dcursors_collection[
            context.scene.C3DB_3Dcursors_index
        ]
        if item.rotation_mode == "QUATERNION":
            rotation = Quaternion(item.rotation_quaternion)
            rotation_euler = rotation.to_euler()
        elif item.rotation_mode == "AXIS_ANGLE":
            rotation = Quaternion(
                Vector(item.rotation_axis_angle).yzw, item.rotation_axis_angle[0]
            )
            rotation_euler = rotation.to_euler()
        else:
            rotation_euler = item.rotation_euler

        bpy.ops.object.empty_add(
            type="PLAIN_AXES", radius=1, location=item.location, rotation=rotation_euler
        )

        bpy.context.object.name = item.name

        return {"FINISHED"}


class C3DB_PT_panel(bpy.types.Panel):
    bl_label = "3D Cursors Briefcase"
    bl_idname = "C3DB_PT_panel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "View"
    bl_parent_id = "VIEW3D_PT_view3d_cursor"

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        col = row.column()
        col.template_list(
            "C3DB_UL_list",
            "",
            context.scene,
            "C3DB_3Dcursors_collection",
            context.scene,
            "C3DB_3Dcursors_index",
            rows=3,
        )
        col = row.column(align=True)
        col.operator("c3db.save", icon="ADD", text="")
        col.operator("c3db.remove_from_list", icon="REMOVE", text="")
        col.separator()
        col.menu("C3DB_MT_menu_specials", icon="DOWNARROW_HLT", text="")
        col.separator()
        if (
            context.scene.C3DB_3Dcursors_collection
            and context.scene.C3DB_3Dcursors_index
            < len(context.scene.C3DB_3Dcursors_collection)
        ):
            col.operator("c3db.move_up_list", icon="TRIA_UP", text="")
            col.operator("c3db.move_down_list", icon="TRIA_DOWN", text="")
        row = layout.row()
        row.operator("c3db.restore", text="Load")
        row.operator("view3d.view_center_cursor", text="Center View")
        row.prop(context.scene, "c3db_auto_center_view_checkbox")


classes = (
    C3DB_PG_properties,
    C3DB_UL_list,
    C3DB_MT_menu_specials,
    C3DB_OT_save,
    C3DB_OT_restore,
    C3DB_OT_update,
    C3DB_OT_restore_next,
    C3DB_OT_restore_previous,
    C3DB_OT_delete_all,
    C3DB_OT_remove_from_list,
    C3DB_OT_move_up_list,
    C3DB_OT_move_down_list,
    C3DB_OT_convert_all_to_empties,
    C3DB_OT_convert_selected_to_empty,
    C3DB_PT_panel,
)

c3db_keymaps = []


def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.Scene.C3DB_3Dcursors_collection = bpy.props.CollectionProperty(
        type=C3DB_PG_properties
    )
    bpy.types.Scene.C3DB_3Dcursors_index = bpy.props.IntProperty()

    bpy.types.Scene.c3db_auto_center_view_checkbox = bpy.props.BoolProperty(
        name="", description="Auto center view on 3D Cursor", default=False
    )

    wm = bpy.context.window_manager
    km = wm.keyconfigs.addon.keymaps.new(name="3D View", space_type="VIEW_3D")

    kmi = km.keymap_items.new(
        C3DB_OT_save.bl_idname, "S", "PRESS", ctrl=True, shift=True
    )
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(
        C3DB_OT_remove_from_list.bl_idname, "D", "PRESS", ctrl=True, shift=True
    )
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(C3DB_OT_restore.bl_idname, "R", "PRESS", ctrl=True)
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(
        C3DB_OT_update.bl_idname,
        "U",
        "PRESS",
        ctrl=True,
        shift=True,
    )
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(
        C3DB_OT_restore_next.bl_idname, "R", "PRESS", ctrl=True, shift=True
    )
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(
        C3DB_OT_restore_previous.bl_idname,
        "R",
        "PRESS",
        ctrl=True,
        shift=True,
        alt=True,
    )
    c3db_keymaps.append((km, kmi))

    kmi = km.keymap_items.new(
        C3DB_OT_delete_all.bl_idname,
        "X",
        "PRESS",
        ctrl=True,
        shift=True,
    )
    c3db_keymaps.append((km, kmi))


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
    del bpy.types.Scene.C3DB_3Dcursors_collection
    del bpy.types.Scene.C3DB_3Dcursors_index

    del bpy.types.Scene.c3db_auto_center_view_checkbox

    wm = bpy.context.window_manager
    for km, kmi in c3db_keymaps:
        km.keymap_items.remove(kmi)
    c3db_keymaps.clear()


if __name__ == "__main__":
    register()
